-> graphical example, showing subsongs/wind, how to do 'toggle-pause',
-> status bar, etc

OPT OSVERSION=37

MODULE	'asl', 'libraries/asl', 'tools/EasyGUI', 'tools/file',
	'*thxplay', 'utility/tagitem', 'plugins/ticker'
-> You should use 'thxplay', not '*thxplay'

DEF mod=NIL, file[2048]:STRING, stopped=TRUE, song=0, gh,
    req=NIL:PTR TO filerequester, title[256]:STRING, status

PROC main() HANDLE
  DEF ticker=NIL:PTR TO ticker

  IF (thxplaybase := OpenLibrary('thxplay.library', 5)) = NIL
    PutStr('Can''t open thxplay.library v5\n')
    Raise()
  ENDIF

  IF ThxInit(NIL)
    PutStr('Can''t initialise THX player\n')
    Raise()
  ENDIF

  IF (aslbase := OpenLibrary('asl.library', 37)) = NIL
    PutStr('Can''t open asl.library v37\n')
    Raise()
  ENDIF

  IF (req := AllocAslRequest(ASL_FILEREQUEST,
    [ASLFR_TITLETEXT, 'Select THX module to play', TAG_DONE]
  )) = NIL THEN Raise()

  status  := 'stopped'

  NEW ticker

  easyguiA('PlayTHX',
    [ROWS,
      [EQCOLS,
        [SBUTTON, {psub}, '|<',    0,   0],
        [SBUTTON, {rew},  '<<',    0,   0],
        [SBUTTON, {play}, '_Play', 0, "p"],
        [SBUTTON, {ffwd}, '>>',    0,   0],
        [SBUTTON, {nsub}, '>|',    0,   0]
      ],
      [EQCOLS,
        [SBUTTON, {stop}, '_Stop',    0, "s"],
        [SBUTTON, {eject},'_Eject',   0, "e"],
        [SBUTTON, {open}, '_Open...', 0, "o"]
      ],
      [PLUGIN, {updatetitle}, ticker]
    ],
    [
      EG_GHVAR, {gh},
      TAG_DONE
    ]
  )

EXCEPT DO
  END ticker

  eject()   -> free module (if loaded)
  IF thxplaybase THEN ThxFree() -> free player

  IF req THEN FreeAslRequest(req)
  CloseLibrary(aslbase)
  CloseLibrary(thxplaybase)
ENDPROC


PROC rew() IS ThxWind(-1)

PROC ffwd() IS ThxWind(1)

PROC psub()
  stop()
  IF song > 0 THEN ThxSetSong(song := song - 1)
  play()
ENDPROC

PROC nsub()
  stop()
  IF song < ThxGetNumSongs() THEN ThxSetSong(song := song + 1)
  play()
ENDPROC

-> doubles as pause()
PROC play()
  IF mod = NIL THEN RETURN

  IF stopped
    ThxSetVolume(64)
    ThxPlay()            -> start playing
    stopped := FALSE     -> no longer stopped
    status  := 'playing'
  ELSE
    -> if playing (not stopped) then pause
    ThxSetVolume(0)
    Delay(2)             -> (setvolume needs 'a while' to happen)
    ThxPause()           -> pause
    stopped := TRUE      -> paused is ~same as stopped :)
    status  := 'paused'
  ENDIF
ENDPROC

PROC stop()
  IF mod = NIL THEN RETURN
  ThxStop()
  stopped := TRUE  -> mark as stopped
  status  := 'stopped'
ENDPROC

PROC eject()
  IF mod = NIL THEN RETURN
  stop()
  freefile(mod)  -> free module
  song    := 0
  stopped := TRUE
  mod     := NIL
  status  := 'stopped'
ENDPROC

PROC open()
  IF AslRequest(req, NIL)
    StrCopy(file, req.drawer)
    AddPart(file, req.file, StrMax(file))
  ELSE
    RETURN -> return if no file chosen
  ENDIF

  eject()                       -> eject current module
  mod := readfile(file)         -> read new module
  IF mod = NIL THEN RETURN      -> no mod loaded? give up
  IF ThxInit(mod) THEN RETURN   -> ThxInit() failed? give up

  play()
ENDPROC

-> this is called every-so-often by the ticker plugin
PROC updatetitle(i,t)
  DEF min, sec

  sec, min := Mod(ThxPlaytime(), 60)

  StringF(title, 'PlayTHX -[\z\d[2]:\z\d[2]]- \s \s [\d/\d] \s',
    min, sec, status, FilePart(file), song + 1, ThxGetNumSongs() + 1,
    IF ThxSongEnded() THEN '*LOOP*' ELSE ''
  )
  changetitle(gh, title)
ENDPROC
